// FORMWHIZ Version 1.0
// Copyright (c) 1999 Ziff-Davis, Inc.
// All rights reserved.
// First Published in PC Magazine, US Edition, September 21, 1999.
// Programmer: Patrick Philippot

// PropPages.cpp : implementation file

#include "stdafx.h"
#include "FWconf.h"
#include "PropPages.h"
#include "DefaultsDlg.h"
#include <atlbase.h>
#include "..\FormWhizKeys.h"
#include "..\EncryptString.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const PTCHAR szFormWhiz		= _T("SOFTWARE\\PC Magazine\\FormWhiz");
const PTCHAR szAssociations	= _T("Associations");

/////////////////////////////////////////////////////////////////////////////
// CAssocPage property page

IMPLEMENT_DYNCREATE(CAssocPage, CPropertyPage)

BEGIN_MESSAGE_MAP(CAssocPage, CPropertyPage)
	//{{AFX_MSG_MAP(CAssocPage)
	ON_LBN_SELCHANGE(IDC_KEYWORDS, OnSelChangeKeywords)
	ON_WM_DESTROY()
	ON_BN_CLICKED(IDC_REMOVE, OnRemoveAssoc)
	ON_BN_CLICKED(IDC_UP, OnUp)
	ON_BN_CLICKED(IDC_DOWN, OnDown)
	ON_BN_CLICKED(IDC_NEW, OnNewAssoc)
	ON_BN_CLICKED(IDC_FORCEFILLIN, OnChangeForceFillin)
	ON_EN_KILLFOCUS(IDC_KEYWORD, NotifyKillFocus)
	ON_EN_KILLFOCUS(IDC_TOKENS, NotifyKillFocus)
	ON_MESSAGE(WM_EDITLOSTFOCUS,OnEditLostFocus)
	ON_BN_CLICKED(IDC_PARTIALMATCH, NotifyKillFocus)
	ON_EN_CHANGE(IDC_TOKENS, OnChangeTokens)
	ON_BN_CLICKED(IDC_EMPTY, OnEmpty)
	ON_BN_CLICKED(IDC_DEFAULTS, OnDefaults)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CAssocPage::CAssocPage() : CPropertyPage(CAssocPage::IDD)
{
	//{{AFX_DATA_INIT(CAssocPage)
	m_bForceFillIn = FALSE;
	//}}AFX_DATA_INIT
	m_pToolTip = NULL;
}

CAssocPage::~CAssocPage()
{
	delete m_pToolTip;
}

void CAssocPage::DoDataExchange(CDataExchange* pDX)
{
	CPropertyPage::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAssocPage)
	DDX_Check(pDX, IDC_FORCEFILLIN, m_bForceFillIn);
	//}}AFX_DATA_MAP
}

BOOL CAssocPage::PreTranslateMessage(MSG* pMsg) 
{
    if (NULL != m_pToolTip)
        m_pToolTip->RelayEvent(pMsg);

    return CPropertyPage::PreTranslateMessage(pMsg);
}

/////////////////////////////////////////////////////////////////////////////
// CAssocPage message handlers

// Format of an association string:
// field_data;y/n;token1;token2;token3;...
// The second field is 'y' if Partial Matches 
// are allowed and 'n' otherwise

BOOL CAssocPage::OnInitDialog() 
{
	CPropertyPage::OnInitDialog();

	// Subclass controls that will be manipulated directly	
	m_listKeywords.SubclassDlgItem(IDC_KEYWORDS, this);
	m_edKeyword.SubclassDlgItem(IDC_KEYWORD, this);
	m_edTokens.SubclassDlgItem(IDC_TOKENS, this);

	//Set up the tooltip
	m_pToolTip = new CToolTipCtrl;
	if(!m_pToolTip->Create(this))
	{
	   TRACE("Unable To create ToolTip\n");
	   return TRUE;
	}

	m_pToolTip->AddTool(&m_listKeywords, IDS_TT_KEYWORDS);
	m_pToolTip->AddTool(&m_edKeyword, IDS_TT_KEYWORD);
	m_pToolTip->AddTool(&m_edTokens, IDS_TT_TOKENS);
	m_pToolTip->AddTool(GetDlgItem(IDC_PARTIALMATCH), IDS_TT_PARTIALMATCH);
	m_pToolTip->AddTool(GetDlgItem(IDC_DEFAULTS), IDS_TT_DEFAULTS);
	m_pToolTip->AddTool(GetDlgItem(IDC_FORCEFILLIN), IDS_TT_FORCEFILLIN);
	m_pToolTip->AddTool(GetDlgItem(IDC_EMPTY), IDS_TT_EMPTY);

	m_pToolTip->Activate(TRUE);

	InitControls();

	return TRUE;
}

void CAssocPage::OnSelChangeKeywords() 
{
	DisplayKeyword(m_listKeywords.GetCurSel());
	SetUpDownButtons();
}

void CAssocPage::OnChangeTokens() 
{
	CString sKeyword;
	CString	sTokens;

	m_edKeyword.GetWindowText(sKeyword);
	m_edTokens.GetWindowText(sTokens);
	if (!sKeyword.IsEmpty() && !sTokens.IsEmpty())
		SetModified();
}


BOOL CAssocPage::OnApply() 
{
	// Make sure that the current association has been updated if needed
	UpdateAssoc(m_listKeywords.GetCurSel());

	CRegKey regkey;

	// Save to m_bForceFillIn flag to registry
	if (regkey.Create(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
	{
		UpdateData(FALSE);
		regkey.SetValue(m_bForceFillIn? 1 : 0, g_szForceFillIn);

		regkey.Close();
	}

	// Delete current data in Association key
	regkey.Open(HKEY_CURRENT_USER, szFormWhiz);
	regkey.RecurseDeleteKey(szAssociations);
	regkey.Close();

	// Re-create association data from list
	int nCount = m_listKeywords.GetCount();
	if ((nCount > 0) && ((regkey.Create(HKEY_CURRENT_USER, g_szAssocKey) == ERROR_SUCCESS)))
	{
		for (int i = 1; i <= nCount; i++)
		{
			CString	sValueName;
			CString	sValue;

			sValueName.Format(_T("%d"), i);
			m_listKeywords.GetText(i - 1, sValue);
			sValue += _T(";");
			sValue +=  (LPSTR) m_listKeywords.GetItemData(i - 1);
			EncryptString(TRUE, sValue.GetBuffer(sValue.GetLength() + 1));
			regkey.SetValue(sValue, sValueName);
		}

		regkey.Close();
	}

	return CPropertyPage::OnApply();
}

void CAssocPage::OnDestroy()
{
	CleanUpList();
	CPropertyPage::OnDestroy();
}

void CAssocPage::OnChangeForceFillin() 
{
	SetModified();
}

LONG CAssocPage::OnEditLostFocus(UINT id, LONG nIndex)
{
	// We need to store the current selection index
	// for the IDC_KEYWORDS listbox because the focus
	// may have been lost by the edit controls because
	// the user clicked on the listbox. In that case,
	// we must preserve the current index which may have
	// been modified by UpdateAssoc().
	//
	// When we reach this code and if id == IDC_KEYWORDS,
	// we may safely assume that one of the edit controls 
	// lost the focus because the user clicked on the 
	// listbox which now possible have have a current
	// selection. This is because we *posted* the
	// notification message.

	int nCurIndex = m_listKeywords.GetCurSel();

	UpdateAssoc(nIndex);

	if ((id == IDC_KEYWORDS) && (nCurIndex != LB_ERR))
		m_listKeywords.SetCurSel(nCurIndex);

	return 0;
}

void CAssocPage::UpdateAssoc(int nIndex)
{
	CString sKeyword;
	CString	sTokens;
	BOOL	bPartialMatch = ((CButton*) GetDlgItem(IDC_PARTIALMATCH))->GetCheck() == 1;

	m_edKeyword.GetWindowText(sKeyword);
	m_edTokens.GetWindowText(sTokens);

	if (!sKeyword.IsEmpty())
	{
		// If the same keyword already exists, use it or display existing information
		// int nCurIndex = m_listKeywords.FindStringExact(-1, sKeyword);

		if (sTokens.IsEmpty())
			return;
		else
		{
			// Remove keyword from list
			if (nIndex != LB_ERR)
			{
				delete (PTCHAR) m_listKeywords.GetItemData(nIndex);
				m_listKeywords.DeleteString(nIndex);
				// Update keyword name
				m_listKeywords.InsertString(nIndex, sKeyword);
			}
			else
				// We probably just clicked on the New button
				// Add the new keyword to the list
				nIndex = m_listKeywords.AddString(sKeyword);

			PTCHAR lpszTokens = new TCHAR[_tcslen(sTokens) + sizeof(TCHAR) + 2];
			_tcscpy(lpszTokens + 2, sTokens);
			*lpszTokens = bPartialMatch ? g_chPartialMatchOKChar : g_chPartialMatchNotOKChar;
			*(lpszTokens+1) = (TCHAR) ';';
			m_listKeywords.SetItemData(nIndex, (DWORD) lpszTokens);

			DisplayKeyword(nIndex);
			SetModified();
		}
	}
}

void CAssocPage::OnNewAssoc() 
{
	// Just prepare the dialog box
	// Then let the UpdateAssoc routine do its job

	CString sKeyword;
	CString	sTokens;

	m_edKeyword.SetSel(0, -1);
	m_edKeyword.Clear();
	m_edTokens.SetSel(0, -1);
	m_edTokens.Clear();

	m_listKeywords.SetCurSel(-1);
	GetDlgItem(IDC_KEYWORD)->SetFocus();
	GetDlgItem(IDC_REMOVE)->EnableWindow(FALSE);
	GetDlgItem(IDC_NEW)->EnableWindow(FALSE);

	SetUpDownButtons();
}

void CAssocPage::OnRemoveAssoc() 
// Remove association from list
{
	int nIndex;

	if ((nIndex = m_listKeywords.GetCurSel()) != LB_ERR)
	{
		delete (PTCHAR) m_listKeywords.GetItemData(nIndex);
		m_listKeywords.DeleteString(nIndex);

		DisplayKeyword(0);
		SetModified();
	}
}

void CAssocPage::OnUp() 
// Manage priority order
{
	int nIndex;

	if (((nIndex = m_listKeywords.GetCurSel()) != LB_ERR)
		&&
		(nIndex != 0))
	{
		CString sText;
		DWORD	dwData = m_listKeywords.GetItemData(nIndex);

		m_listKeywords.GetText(nIndex, sText);
		m_listKeywords.DeleteString(nIndex);

		int nNewIndex = nIndex - 1;
		m_listKeywords.InsertString(nNewIndex, sText);
		m_listKeywords.SetItemData(nNewIndex, dwData);
		m_listKeywords.SetCurSel(nNewIndex);

		SetModified();
	}	

	SetUpDownButtons();
}

void CAssocPage::OnDown() 
// Manage priority order
{
	int nIndex;

	if (((nIndex = m_listKeywords.GetCurSel()) != LB_ERR)
		&&
		(nIndex < m_listKeywords.GetCount() - 1))
	{
		CString sText;
		DWORD	dwData = m_listKeywords.GetItemData(nIndex);

		m_listKeywords.GetText(nIndex, sText);
		m_listKeywords.DeleteString(nIndex);

		int nNewIndex = nIndex + 1;
		m_listKeywords.InsertString(nNewIndex, sText);
		m_listKeywords.SetItemData(nNewIndex, dwData);
		m_listKeywords.SetCurSel(nNewIndex);

		SetModified();
	}	

	SetUpDownButtons();
}	

void CAssocPage::SetUpDownButtons()
{
	int nIndex;

	if ((nIndex = m_listKeywords.GetCurSel()) != LB_ERR)
	{
		GetDlgItem(IDC_UP)->EnableWindow(nIndex > 0);
		GetDlgItem(IDC_DOWN)->EnableWindow(nIndex < m_listKeywords.GetCount() - 1);
	}
	else
	{
		GetDlgItem(IDC_UP)->EnableWindow(FALSE);
		GetDlgItem(IDC_DOWN)->EnableWindow(FALSE);
	}
}

void CAssocPage::InitControls()
{
	CRegKey regkey;

	// Get the value of m_bForceFillIn from registry
	if (regkey.Open(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
	{
		DWORD	dwForceFillIn;

		regkey.QueryValue(dwForceFillIn, g_szForceFillIn);
		m_bForceFillIn = dwForceFillIn == 1;

		UpdateData(FALSE);

		regkey.Close();
	}

	m_edKeyword.SetLimitText(MAX_FILL);
	m_edTokens.SetLimitText(MAX_TOKENS);

	// Get all associations
	if (regkey.Open(HKEY_CURRENT_USER, g_szAssocKey) == ERROR_SUCCESS)
	{
		TCHAR 	szBuffer[MAX_FILL + MAX_TOKENS];
		TCHAR 	szValueName[MAX_NAME];
		int		nIndex = 1;
		DWORD	dwBufSize = sizeof(szBuffer);

		_itot(nIndex, szValueName, 10);

		// Read each association
		while (regkey.QueryValue(szBuffer, szValueName, &dwBufSize) == ERROR_SUCCESS)
		{
			EncryptString(FALSE, szBuffer);

			PTCHAR pPos = _tcschr(szBuffer, (TCHAR) ';');

			if (pPos)
			{
				*pPos = (TCHAR) '\0';
				pPos++;

				int nItemIndex = m_listKeywords.AddString(szBuffer);

				PTCHAR pTokens = new TCHAR[_tcslen(pPos) + sizeof(TCHAR)];
				_tcscpy(pTokens, pPos);
				m_listKeywords.SetItemData(nItemIndex, (DWORD) pTokens);
			}

			nIndex++;
			_itot(nIndex, szValueName, 10);
			dwBufSize = sizeof(szBuffer);
		}

		DisplayKeyword(0);
		SetUpDownButtons();
		regkey.Close();
	}
}

void CAssocPage::CleanUpList()
// Delete associated token strings before closing
// This to prevent from creating memory leaks.
{
	int nCount = m_listKeywords.GetCount();

	for (int i = 0; i < nCount; i++)
		delete (PTCHAR) m_listKeywords.GetItemData(i);
}

void CAssocPage::DisplayKeyword(int nIndex)
// Display specified item
{
	TCHAR	szFill[MAX_FILL];
	PTCHAR	pszTokens;

	if (m_listKeywords.GetText(nIndex, szFill) != LB_ERR)
	{
		m_edKeyword.SetWindowText(szFill);
		pszTokens = (LPSTR) m_listKeywords.GetItemData(nIndex);
		m_edTokens.SetWindowText(pszTokens + 2);
		m_listKeywords.SetCurSel(nIndex);
		((CButton*) GetDlgItem(IDC_PARTIALMATCH))->SetCheck(*pszTokens == g_chPartialMatchOKChar ? 1 : 0);
		GetDlgItem(IDC_REMOVE)->EnableWindow(TRUE);
		GetDlgItem(IDC_NEW)->EnableWindow(TRUE);
	}
}

// The code below is necessary because we want to update
// associations as soon as the Edit controls lose the
// focus (which makes an Update button unnecessary).
// However, we can't process this update from within
// the EN_KILLFOCUS message handler. So, we *post* a private
// message and handle it in OnEditLostFocus. Sending a
// synchronous message wouldn't help.

void CAssocPage::NotifyKillFocus()
{
	CWnd* pWnd = GetFocus();

	if (pWnd)
		PostMessage(WM_EDITLOSTFOCUS, pWnd->GetDlgCtrlID(), (LPARAM) m_listKeywords.GetCurSel());
}	

void CAssocPage::OnEmpty() 
{
	m_edKeyword.SetWindowText(g_szEmpty);
	UpdateAssoc(m_listKeywords.GetCurSel());
	m_edTokens.SetFocus();
}

void CAssocPage::OnDefaults() 
{
	CDefaultsDlg dlg;

	if (dlg.DoModal() == IDOK)
	{
		int nPos = dlg.m_sDefaults.Find((TCHAR) ';');
		if (nPos != -1)
		{
			((CButton*) GetDlgItem(IDC_PARTIALMATCH))->SetCheck(dlg.m_sDefaults.GetAt(0) == g_chPartialMatchOKChar ? 1 : 0);
			m_edTokens.SetWindowText(dlg.m_sDefaults.Mid(nPos + 1));
			m_edTokens.SetFocus();
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// CKeyPage property page

IMPLEMENT_DYNCREATE(CKeyPage, CPropertyPage)

BEGIN_MESSAGE_MAP(CKeyPage, CPropertyPage)
	//{{AFX_MSG_MAP(CKeyPage)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CKeyPage::CKeyPage() : CPropertyPage(CKeyPage::IDD)
{
	//{{AFX_DATA_INIT(CKeyPage)
	//}}AFX_DATA_INIT
	m_pToolTip = NULL;
}

CKeyPage::~CKeyPage()
{
	delete m_pToolTip;
}

void CKeyPage::DoDataExchange(CDataExchange* pDX)
{
	CPropertyPage::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CKeyPage)
	//}}AFX_DATA_MAP
}

BOOL CKeyPage::PreTranslateMessage(MSG* pMsg) 
{
    if (NULL != m_pToolTip)
        m_pToolTip->RelayEvent(pMsg);

	return CPropertyPage::PreTranslateMessage(pMsg);
}

/////////////////////////////////////////////////////////////////////////////
// CKeyPage message handlers

BOOL CKeyPage::OnInitDialog() 
{
	CPropertyPage::OnInitDialog();
	
	m_keySerialize.SubclassDlgItem(IDC_HOTKEY_RECORD, this);
	m_keyDelete.SubclassDlgItem(IDC_HOTKEY_DELETE, this);
	m_keyConfig.SubclassDlgItem(IDC_HOTKEY_CONFIG, this);
	m_keyToggle.SubclassDlgItem(IDC_HOTKEY_TOGGLE, this);

	//Set up the tooltip
	m_pToolTip = new CToolTipCtrl;
	if(!m_pToolTip->Create(this))
	{
	   TRACE("Unable To create ToolTip\n");
	   return TRUE;
	}

	m_pToolTip->AddTool(GetDlgItem(IDC_HOTKEY_RECORD), IDS_TT_HKRECORD);
	m_pToolTip->AddTool(GetDlgItem(IDC_HOTKEY_DELETE), IDS_TT_HKDELETE);
	m_pToolTip->AddTool(GetDlgItem(IDC_HOTKEY_CONFIG), IDS_TT_HKCONFIG);
	m_pToolTip->AddTool(GetDlgItem(IDC_HOTKEY_TOGGLE), IDS_TT_HKTOGGLE);

	InitControls();
	
	return TRUE;
}

BOOL CKeyPage::OnApply() 
{
	CRegKey regkey;

	// Open or Create the hotkeys
	if (regkey.Create(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
	{
		DWORD dwKey;

		dwKey = m_keySerialize.GetHotKey();
		regkey.SetValue(dwKey, g_szSerializeKey);
		dwKey = m_keyDelete.GetHotKey();
		regkey.SetValue(dwKey, g_szDeleteKey);
		dwKey = m_keyConfig.GetHotKey();
		regkey.SetValue(dwKey, g_szConfigKey);
		dwKey = m_keyToggle.GetHotKey();
		regkey.SetValue(dwKey, g_szToggleKey);

		regkey.Close();
	}
	
	return CPropertyPage::OnApply();
}

void CKeyPage::InitControls()
{
	CRegKey regkey;

	// Open or Create the DeleteKey and SerializeKey keys
	if (regkey.Open(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
	{
		DWORD dwKey;

		if (regkey.QueryValue(dwKey, g_szSerializeKey) == ERROR_SUCCESS)
			m_keySerialize.SetHotKey(LOBYTE(dwKey), HIBYTE(dwKey));
			
		if (regkey.QueryValue(dwKey, g_szDeleteKey) == ERROR_SUCCESS)
			m_keyDelete.SetHotKey(LOBYTE(dwKey), HIBYTE(dwKey));
			
		if (regkey.QueryValue(dwKey, g_szConfigKey) == ERROR_SUCCESS)
			m_keyConfig.SetHotKey(LOBYTE(dwKey), HIBYTE(dwKey));

		if (regkey.QueryValue(dwKey, g_szToggleKey) == ERROR_SUCCESS)
			m_keyToggle.SetHotKey(LOBYTE(dwKey), HIBYTE(dwKey));

		regkey.Close();
	}
}

/////////////////////////////////////////////////////////////////////////////
// CURLPage property page

IMPLEMENT_DYNCREATE(CURLPage, CPropertyPage)

BEGIN_MESSAGE_MAP(CURLPage, CPropertyPage)
	//{{AFX_MSG_MAP(CURLPage)
	ON_LBN_SELCHANGE(IDC_URLS, OnSelchangeUrls)
	ON_BN_CLICKED(IDC_REMOVE, OnRemove)
	ON_WM_DESTROY()
	ON_BN_CLICKED(IDC_REFRESH, InitControls)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

CURLPage::CURLPage() : CPropertyPage(CURLPage::IDD)
{
	//{{AFX_DATA_INIT(CURLPage)
	//}}AFX_DATA_INIT
	m_pToolTip = NULL;
}

CURLPage::~CURLPage()
{
	delete m_pToolTip;
}

void CURLPage::DoDataExchange(CDataExchange* pDX)
{
	CPropertyPage::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CURLPage)
	//}}AFX_DATA_MAP
}

BOOL CURLPage::PreTranslateMessage(MSG* pMsg) 
{
    if (NULL != m_pToolTip)
        m_pToolTip->RelayEvent(pMsg);

	return CPropertyPage::PreTranslateMessage(pMsg);
}

/////////////////////////////////////////////////////////////////////////////
// CURLPage message handlers

BOOL CURLPage::OnInitDialog() 
{
	CPropertyPage::OnInitDialog();
	
	m_listURLs.SubclassDlgItem(IDC_URLS, this);
	m_stTarget.SubclassDlgItem(IDC_TARGET_URL, this);

	//Set up the tooltip
	m_pToolTip = new CToolTipCtrl;
	if(!m_pToolTip->Create(this))
	{
	   TRACE("Unable To create ToolTip\n");
	   return TRUE;
	}

	m_pToolTip->AddTool(GetDlgItem(IDC_URLS), IDS_TT_URLS);
	m_pToolTip->AddTool(GetDlgItem(IDC_REMOVE), IDS_TT_REMOVE);
	m_pToolTip->AddTool(GetDlgItem(IDC_REFRESH), IDS_TT_REFRESH);
	
	InitControls();
	
	return TRUE;
}

void CURLPage::OnSelchangeUrls() 
{
	DisplayURL(m_listURLs.GetCurSel());
	GetDlgItem(IDC_REMOVE)->EnableWindow(m_listURLs.GetCount() != 0);
}

void CURLPage::OnDestroy() 
{
	CleanUpList();
	CPropertyPage::OnDestroy();
}

void CURLPage::OnRemove() 
// Destroy an entry in the recorded URLs list
{
	CRegKey regkey;

	if (regkey.Open(HKEY_CURRENT_USER, g_szURLKey) == ERROR_SUCCESS)
	{
		int nIndex = m_listURLs.GetCurSel();

		if (nIndex != LB_ERR)
		{
			CString sTitle;
			CString sFullTitle;

			// Get title
			m_listURLs.GetText(nIndex, sTitle);
			sFullTitle = g_szFWPrefix;
			sFullTitle += sTitle;
			
			CString sURL;
			DWORD	dwSize = MAX_URL * sizeof(TCHAR);

			// Get corresponding registry entry
			regkey.QueryValue(sURL.GetBuffer(MAX_URL), sFullTitle, &dwSize);
			sURL.ReleaseBuffer();

			CString sFilename;
			dwSize = MAX_PATH * sizeof(TCHAR);

			// Get name of associated form file
			regkey.QueryValue(sFilename.GetBuffer(MAX_PATH), sURL, &dwSize);
			sFilename.ReleaseBuffer();

			if ((!sURL.IsEmpty()) && (!sFullTitle.IsEmpty()))
			{
				// Cleanup registry
				regkey.DeleteValue(sURL);
				regkey.DeleteValue(sFullTitle);

				// Destroy file
				DeleteFile(sFilename);
				// Cleanup list
				m_listURLs.DeleteString(nIndex);
				m_listURLs.SetCurSel(0);
			}
		}

		regkey.Close();

		GetDlgItem(IDC_REMOVE)->EnableWindow(m_listURLs.GetCount() != 0);
	}
}

void CURLPage::InitControls()
{
	CRegKey regkey;

	m_listURLs.ResetContent();

	if (regkey.Open(HKEY_CURRENT_USER, g_szURLKey) == ERROR_SUCCESS)
	{
		DWORD	dwIndex = 0;
		DWORD	dwType;
		TCHAR	szTitle[MAX_URL];
		DWORD	dwValueSize = sizeof(szTitle);
		BYTE	URLBuf[MAX_URL];
		DWORD	dwDataSize = sizeof(URLBuf);

		while (::RegEnumValue(regkey.m_hKey, dwIndex, szTitle, &dwValueSize, NULL, &dwType, URLBuf, &dwDataSize) == ERROR_SUCCESS)
		{
			if ((dwType == REG_SZ) && (!_tcsnicmp(szTitle, g_szFWPrefix, _tcslen(g_szFWPrefix))))
			// We have a title entry (we skip other entries)
			{
				int nItemIndex = m_listURLs.AddString(szTitle + _tcslen(g_szFWPrefix));

				PTCHAR pURL = new TCHAR[dwDataSize];
				if (pURL)
				{
					_tcscpy(pURL, (LPSTR) URLBuf);
					m_listURLs.SetItemData(nItemIndex, (DWORD) pURL);
				}
			}

			dwValueSize = sizeof(szTitle);
			dwDataSize = sizeof(URLBuf);
			dwIndex++;
		}

		BOOL bOK = m_listURLs.GetCount() != 0;
		if (bOK)
			DisplayURL(0);
		GetDlgItem(IDC_REMOVE)->EnableWindow(bOK);

		regkey.Close();
	}
}

void CURLPage::DisplayURL(int nIndex)
{
	TCHAR szURL[MAX_URL];

	if (m_listURLs.GetText(nIndex, szURL) != LB_ERR)
	{
		m_stTarget.SetWindowText((LPSTR) m_listURLs.GetItemData(nIndex));
		m_listURLs.SetCurSel(nIndex);
	}
}

void CURLPage::CleanUpList()
{
	int nCount = m_listURLs.GetCount();

	for (int i = 0; i < nCount; i++)
		delete (PTCHAR) m_listURLs.GetItemData(i);
}
