// FORMWHIZ Version 1.0
// Copyright (c) 1999 Ziff-Davis, Inc.
// All rights reserved.
// First Published in PC Magazine, US Edition, September 21, 1999.
// Programmer: Patrick Philippot

// FormWhizObj.cpp : Implementation of CFormWhizObj

#include "stdafx.h"
#include <tchar.h>
#include <ExDisp.h>
#include <ExDispID.h>
#include <comdef.h>
#include <Commctrl.h>
#include "regexp.h"

#include "uistr.h"
#include "FormWhizKeys.h"
#include "EncryptString.h"
#include "FormWhiz.h"
#include "FormWhizObj.h"


TCHAR const g_szAppName[] 		= _T("FormWhiz");
TCHAR const g_szModuleName[] 	= _T("FormWhiz.dll");
TCHAR const g_szConfName[]	 	= _T("FWConf.exe");
TCHAR const g_szData[]		 	= _T("Data");
TCHAR const g_szForm[]		 	= _T("Form");
TCHAR const g_szNull[]	 		= _T("");
WCHAR const g_szNullW[] 		= L"";


// HTML related strings
WCHAR const g_szHtmlType[] 		= L"TYPE";
WCHAR const g_szHtmlName[] 		= L"NAME";
WCHAR const g_szHtmlValue[]		= L"VALUE";
WCHAR const g_szHtmlEmptyTitle[]= L"EMPTY_OR_ALREADY_USED_TITLE";
WCHAR const g_szHtmlMaxLen[]	= L"MAXLENGTH";
TCHAR const g_szHtmlText[] 		= _T("TEXT");
TCHAR const g_szHtmlInput[]		= _T("INPUT");
TCHAR const g_szHtmlPassword[]	= _T("PASSWORD");
TCHAR const g_szHtmlButton[]	= _T("BUTTON");
TCHAR const g_szHtmlRadio[]		= _T("RADIO");
TCHAR const g_szHtmlCheckBox[]	= _T("CHECKBOX");
TCHAR const g_szHtmlSelect[]	= _T("SELECT");
TCHAR const g_szHtmlOption[]	= _T("OPTION");

#pragma data_seg( "FWObject_Shared" )
BOOL g_bActive = TRUE;
BOOL g_bToggleInUse = FALSE;
int	 g_nInstances = 0;
#pragma data_seg()


// Global mechanism managing associations between CFormWhizObj* and window handles
// Since the keyboard hook is a static method, we need a way to know which
// object instance works with which browser window. Using a global hook as
// is usually done under such circumstances wouldn't work with IE. So we store
// the handle/pointer/thread id associations in a global array of FW_INSTANCE structures.

// There's a simpler way to do that : storing the CFormWhiz pointer with the
// browser window by using SetWindowLong and GWL_USERDATA. However, this method is
// unsafe because other applications (including IE itself) may use this offset.

#define MAX_FW_INSTANCES	100

struct FW_INSTANCE
{
	HWND			hWnd;
	CFormWhizObj*	pFw;
	DWORD			dwThreadId;
} 
FwInstances[MAX_FW_INSTANCES];

int	g_InstanceCount = 0;

BOOL AddInstance(HWND hWnd, CFormWhizObj* pFw, DWORD dwThreadId)
{
	if (g_InstanceCount < MAX_FW_INSTANCES)
	{
		FwInstances[g_InstanceCount].hWnd = hWnd;
		FwInstances[g_InstanceCount].pFw = pFw;
		FwInstances[g_InstanceCount].dwThreadId = dwThreadId;
		g_InstanceCount++;
		return TRUE;
	}
	else
		return FALSE;
}

void RemoveInstance(CFormWhizObj* pFw)
{
	int i;
	for (i = 0; i < g_InstanceCount; i++)
		if (FwInstances[i].pFw == pFw)
		{
			if (i < MAX_FW_INSTANCES - 1)
				memmove(&FwInstances[i], &FwInstances[i+1], sizeof(FW_INSTANCE) * (g_InstanceCount - i - 1));
			g_InstanceCount--;
			break;
		}

	return;
}

CFormWhizObj* LookupInstance(HWND hWnd)
{
	int i;
	for (i = 0; i < g_InstanceCount; i++)
	{
		if (FwInstances[i].hWnd == hWnd)
			return FwInstances[i].pFw;
	}

	return NULL;
}

CFormWhizObj* LookupInstance(DWORD dwThreadId)
{
	int i;
	for (i = 0; i < g_InstanceCount; i++)
	{
		if (FwInstances[i].dwThreadId == dwThreadId)
			return FwInstances[i].pFw;
	}

	return NULL;
}

/////////////////////////////////////////////////////////////////////////////
// CFormWhizObj

CFormWhizObj::CFormWhizObj()
{
	m_bHotkeysEnabled = FALSE;
	m_bDocComplete = FALSE;
	m_nDocNum = 0;
	m_hHook = NULL;

	if (g_nInstances == 0)	// Get active / inactive state from registry
	{
		CRegKey regkey;

		if (regkey.Open(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
		{
			DWORD dwActive;

			if (regkey.QueryValue(dwActive, g_szActive) == ERROR_SUCCESS)
				g_bActive = (BOOL) dwActive;
				
			regkey.Close();
		}
	}
	// Increment instance counter
	// We'll store the current active / inactive state
	// to the registry only when the last instance shuts down
	g_nInstances++;
}

CFormWhizObj::~CFormWhizObj()
{
	g_nInstances--;
	if (g_nInstances == 0)	// Last instance
	{						// Store active / inactive state to registry
		CRegKey regkey;

		if (regkey.Create(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
		{
			regkey.SetValue(g_bActive? 1 : 0, g_szActive);
			regkey.Close();
		}
	}
}

BOOL CFormWhizObj::ManageConnection(enum AdviseAction action)
// Setup advise sink
{
	if (!m_spWebBrowser2)
		return S_OK;

	HRESULT hr;

	CComQIPtr<IConnectionPointContainer, &IID_IConnectionPointContainer> spCPContainer(m_spWebBrowser2);

	if (spCPContainer != NULL)
	{
		CComPtr<IConnectionPoint> spConnectionPoint;

		hr = spCPContainer->FindConnectionPoint(DIID_DWebBrowserEvents2, &spConnectionPoint);
		if (SUCCEEDED(hr))
		{
			if (action == Advise)
				// Advise the client site that we want to sink events
				hr = spConnectionPoint->Advise((IDispatch*)this, &m_dwCookie);
			else
				// Remove us from the list of active sinks
				hr = spConnectionPoint->Unadvise(m_dwCookie);
		}
	}
	
	return (SUCCEEDED(hr));
}

// IOleObjectWithSite Methods
STDMETHODIMP CFormWhizObj::SetSite(IUnknown *pUnkSite)
// Connect to browser
{
	if (pUnkSite != NULL)
	{
		// Query pUnkSite for the IWebBrowser2 interface.
		m_spWebBrowser2 = pUnkSite;

		if (m_spWebBrowser2)
		{
			m_spWebBrowser2->get_HWND((long*) &m_hwndIE);

			// We'll need the this pointer later when the keyboard
			// hook will be called. 
			if (!AddInstance(m_hwndIE, this, GetCurrentThreadId()))
				FormWhizError(IDS_ERROR_TOOMUCHWINDOWS, errError);

			// Connect to the browser in order to handle events.
			if (!ManageConnection(Advise))
				FormWhizError(IDS_ERROR_SINKINGFAILED, errError);
		}

		// Set the keyboard hook
		m_hHook = SetWindowsHookEx(WH_KEYBOARD, 
								   reinterpret_cast<HOOKPROC>(IEKeyboardProc), 
								   NULL, 
								   GetCurrentThreadId());
	}

	return S_OK;
}

// IDispatch Methods
STDMETHODIMP CFormWhizObj::Invoke(DISPID dispidMember, REFIID riid, LCID lcid, WORD wFlags,
                                DISPPARAMS* pDispParams, VARIANT* pvarResult,
                                EXCEPINFO*  pExcepInfo,  UINT* puArgErr)
// Through this dispatch interface, we'll receive all browser events
// Event arguments are listed before each case statement								
{
	if (!pDispParams)
		return E_INVALIDARG;

	CComPtr<IDispatch> spDisp;
	HRESULT hr;

	switch (dispidMember)
	{
		// [0]: Cancel flag  - VT_BYREF|VT_BOOL
		// [1]: HTTP headers - VT_BYREF|VT_VARIANT
		// [2]: Address of HTTP POST data  - VT_BYREF|VT_VARIANT 
		// [3]: Target frame name - VT_BYREF|VT_VARIANT 
		// [4]: Option flags - VT_BYREF|VT_VARIANT
		// [5]: URL to navigate to - VT_BYREF|VT_VARIANT
		// [6]: An object that evaluates to the top-level or frame
		//      WebBrowser object corresponding to the event. 
		//
		case DISPID_BEFORENAVIGATE2:
			// Deactivate keyboard hook
			m_bHotkeysEnabled = FALSE;
			// Do not allow form serialization until document is complete
			m_bDocComplete = FALSE;
			// This is needed for document with Frames
			// If the document contains frames, we'll receive
			// this event for each frame
			m_spWebBrowser2 = pDispParams->rgvarg[6].pdispVal;

			// Get the WebBrowser's document object
			hr = m_spWebBrowser2->get_Document(&spDisp);
			if (SUCCEEDED(hr))
			{
				// Is this an IHTMLDocument2 interface?
				CComQIPtr<IHTMLDocument2, &IID_IHTMLDocument2> spHTML;
				spHTML = spDisp;

				if (spHTML)
					RemoveDocument(spHTML);
			}

			break;
		
		// [0]: URL navigated to - VT_BYREF|VT_VARIANT
		// [1]: An object that evaluates to the top-level or frame
		//      WebBrowser object corresponding to the event. 
		//
		case DISPID_NAVIGATECOMPLETE2:
			m_bHotkeysEnabled = TRUE;
			break;
		
		// [0]: New status bar text - VT_BSTR
		//
		/*
		case DISPID_STATUSTEXTCHANGE:
			break;
		*/
		
		// [0]: Maximum progress - VT_I4
		// [1]: Amount of total progress - VT_I4
		//
		/*
		case DISPID_PROGRESSCHANGE:
			break;
		*/
		
		// [0]: Document URL - VT_BYREF|VT_VARIANT
		// [1]: An object that evaluates to the top-level or frame
		//      WebBrowser object corresponding to the event. 
		case DISPID_DOCUMENTCOMPLETE:
			// Temporarily disable keyboard hook
			m_bHotkeysEnabled = FALSE;

			// Fill in form if necessary
			m_spWebBrowser2 = pDispParams->rgvarg[1].pdispVal;

			// Get the WebBrowser's document object
			hr = m_spWebBrowser2->get_Document(&spDisp);
			if (SUCCEEDED(hr))
			{
				// Verify that what we get is a pointer to a IHTMLDocument2 
				// interface. To be sure, let's query for 
				// the IHTMLDocument2 interface (through smart pointers)
				CComQIPtr<IHTMLDocument2, &IID_IHTMLDocument2> spHTML;
				spHTML = spDisp;
				// The previous step is sufficient to keep Explorer aside
				// without an explicit check against the loader module

				if (spHTML && DoFillIn(spHTML))
					AddDocument(spHTML);
			}
			
			// Activate keyboard hook
			m_bHotkeysEnabled = TRUE;
			m_bDocComplete = TRUE;

			break;

		// No parameters
		/*
		case DISPID_DOWNLOADBEGIN:
			break;
		*/
		
		// No parameters
		/*
		case DISPID_DOWNLOADCOMPLETE:
			break;
		*/
		
		// [0]: Enabled state - VT_BOOL
		// [1]: Command identifier - VT_I4
		//
		/*
		case DISPID_COMMANDSTATECHANGE:
			break;
		*/
		
		// [0]: Address of cancel flag - VT_BYREF|VT_BOOL
		//
		/*
		case DISPID_NEWWINDOW2:
			break;
		*/
		
		// [0]: Document title - VT_BSTR
		// [1]: An object that evaluates to the top-level or frame
		//      WebBrowser object corresponding to the event. 
		//
		/*
		case DISPID_TITLECHANGE:
			break;
		*/

		// [0]: Name of property that changed - VT_BSTR
		//
		/*
		case DISPID_PROPERTYCHANGE:
			break;
		*/

		// [0]: Address of cancel flag - VT_BYREF|VT_BOOL
		//
		case DISPID_ONQUIT:
			// Unregister the keyboard hook previously installed
			// for this instance.
			if (m_hHook)
				UnhookWindowsHookEx(m_hHook);
			ManageConnection(Unadvise);
			RemoveInstance(this);
			break;

		default:
			break;
	}

	return S_OK;
}

BOOL CFormWhizObj::DoFillIn(IHTMLDocument2* pIDoc)
// Does the actual job of filling in a form
{
	CComPtr<IDispatch> spDisp;
	HRESULT hr;
	BOOL	bDone = FALSE;
	BOOL	bForceFillIn = TRUE;

	if (!g_bActive)
		return(TRUE);

	if (!pIDoc)
		return(FALSE);

	// First try to get info from stored form
	// for this URL
	if (SerializeForms(FALSE, pIDoc))
		return(TRUE);
									
	// No recorded form for this URL
	// Get value of ForceFillIn option
	{
		CRegKey regkey;

		if (regkey.Open(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
		{
			DWORD dwForceFillIn;

			if (regkey.QueryValue(dwForceFillIn, g_szForceFillIn) == ERROR_SUCCESS)
				bForceFillIn = (BOOL) dwForceFillIn;
				
			regkey.Close();
		}
	}

	// Start scanning of document
	// Get pointer to Forms collection
	CComPtr<IHTMLElementCollection> pForms;
	long lFormNum, lFormIndex;

	hr = pIDoc->get_forms(&pForms);
	if (SUCCEEDED(hr))
	{
		// Get number of forms
		pForms->get_length(&lFormNum);

		CRegKey regkey;

		if (regkey.Open(HKEY_CURRENT_USER, g_szAssocKey) != ERROR_SUCCESS)
			return(FALSE);

		// For each form
		for (lFormIndex = 0; lFormIndex < lFormNum; lFormIndex++)
		{
			CComQIPtr<IHTMLFormElement, &IID_IHTMLFormElement> spForm;
			spDisp = NULL;

			_variant_t	vName(lFormIndex);
			_variant_t	vIndex;

			// Get next <FORM> tag
			hr = pForms->item(vName, vIndex, &spDisp);
			spForm = spDisp;

			if (SUCCEEDED(hr))
			{
				CComQIPtr<IHTMLElementCollection, &IID_IHTMLElementCollection> spInputElems;
				spDisp = NULL;

				// Get <INPUT> elements collection
				_variant_t vTagName(g_szHtmlInput);
				hr = spForm->tags(vTagName, &spDisp);
				spInputElems = spDisp;

				if (SUCCEEDED(hr))
				{
					long lElemNum, lElemIndex;

					spInputElems->get_length(&lElemNum);
										
					// For each INPUT element of type TEXT or PASSWORD
					for (lElemIndex = 0; lElemIndex < lElemNum; lElemIndex++)
					{
						CComQIPtr<IHTMLElement, &IID_IHTMLElement> spInputElem;
						spDisp = NULL;

						_variant_t	vElemName(lElemIndex);
						_variant_t	vElemIndex;

						// Get element
						hr = spInputElems->item(vElemName, vElemIndex, &spDisp);
						spInputElem = spDisp;

						if (SUCCEEDED(hr))
						{
							BSTR bsAttr;
							_variant_t vValue(g_szNull);
							
							if (!bForceFillIn)
							{
								bsAttr = SysAllocString(g_szHtmlValue);
								// Get element value
								spInputElem->getAttribute(bsAttr, FALSE, &vValue);
								if ((vValue.bstrVal != NULL) && (wcslen((bstr_t) vValue) != 0))
								{
									SysFreeString(bsAttr);
									continue;
								}
								else
									SysReAllocString(&bsAttr, g_szHtmlType);
							}
							else
								bsAttr = SysAllocString(g_szHtmlType);

							// Get element type
							spInputElem->getAttribute(bsAttr, FALSE, &vValue);
							if (!(bstr_t) vValue)
								// No type means TEXT
								vValue.SetString(g_szHtmlText);
							SysFreeString(bsAttr);
							if (!(_tcsicmp((_bstr_t) vValue, g_szHtmlText))
								||
								!(_tcsicmp((_bstr_t) vValue, g_szHtmlPassword)))
							{
								BSTR bsTagName = SysAllocString(g_szHtmlName);
			
								// We have at least one form and one relevant 
								// element in that form
								bDone = TRUE;
								// Get element name
								spInputElem->getAttribute(bsTagName, FALSE, &vValue);
								if (!(bstr_t) vValue)
									vValue.SetString(g_szNull);
								SysFreeString(bsTagName);
								// Compare with lists of possible names
								// If name matches, replace element VALUE with pre-defined text
								if (TagMatch(&vValue, regkey))
								{
									// Get MAXLENGTH attribute
									BSTR bsMaxLen = SysAllocString(g_szHtmlMaxLen);
									_variant_t vMaxLen(g_szNull);
									
									if (spInputElem->getAttribute(bsMaxLen, FALSE, &vMaxLen) == S_OK)
									{
										// User-defined string longer than MAXLENGTH?
										if (wcslen((_bstr_t) vValue) > (UINT) (long) vMaxLen)
										{
											// We use char because SetString doesn't
											// accept a LPSTR, only a char*
											char szBuffer [MAX_FILL];
											
											_uiscpy(szBuffer, (WCHAR*) (_bstr_t) vValue);
											szBuffer[(long) vMaxLen] = '\0';
											vValue.SetString(szBuffer);
										}
									}
									
									bsAttr = SysAllocString(L"VALUE");
									spInputElem->setAttribute(bsAttr, vValue, FALSE);
									SysFreeString(bsAttr);
								}
							}
						}
					}
				}

				// Now enumerate SELECT tags
				CComQIPtr<IHTMLElementCollection, &IID_IHTMLElementCollection> spSelectElems;
				spDisp = NULL;

				// Get <SELECT> element collection
				_variant_t vSelectName(g_szHtmlSelect);
				hr = spForm->tags(vSelectName, &spDisp);
				spSelectElems = spDisp;

				if (SUCCEEDED(hr))
				{
					long lSelectNum, lSelectIndex;

					spSelectElems->get_length(&lSelectNum);
					// For each SELECT element
					for (lSelectIndex = 0; lSelectIndex < lSelectNum; lSelectIndex++)
					{
						CComQIPtr<IHTMLSelectElement, &IID_IHTMLSelectElement> spSelectElem;
						spDisp = NULL;

						_variant_t	vSelectName(lSelectIndex);
						_variant_t	vSelectIndex;

						// get next select element
						hr = spSelectElems->item(vSelectName, vSelectIndex, &spDisp);
						spSelectElem = spDisp;

						if (SUCCEEDED(hr))
						{
							BSTR bsName = SysAllocString(g_szNullW);

							// We have at least one form and one relevant 
							// element in that form
							bDone = TRUE;
							// Get element name
							spSelectElem->get_name(&bsName);
							// Compare with lists of possible names
							// If name matches, replace element VALUE with pre-defined text
							_variant_t	vValue(bsName);
							SysFreeString(bsName);

							if (TagMatch(&vValue, regkey))
							{
								CComQIPtr<IHTMLElementCollection, &IID_IHTMLElementCollection> spOptions;
								spDisp = NULL;

								_variant_t vOptionName(g_szHtmlOption);

								// Get option collection
								hr = spSelectElem->tags(vOptionName, &spDisp);
								spOptions = spDisp;

								if (SUCCEEDED(hr))
								{
									long lOptionNum, lOptionIndex;

									WCHAR wsText[MAX_FILL];
#ifndef _UNICODE
									MultiByteToWideChar(CP_ACP,
														MB_PRECOMPOSED,
														(_bstr_t) vValue,
														_tcslen((_bstr_t) vValue) * sizeof(TCHAR),
														wsText,
														sizeof(wsText) / sizeof (TCHAR));
									wsText[_tcslen((_bstr_t) vValue)]= (WCHAR) '\0';
#else
									_tcscpy(wsText, (_bstr_t) vValue);
#endif
									spOptions->get_length(&lOptionNum);
									for (lOptionIndex = 0; lOptionIndex < lOptionNum; lOptionIndex++)
									{
										CComQIPtr<IHTMLOptionElement, &IID_IHTMLOptionElement>pOption;
										spDisp = NULL;

										_variant_t	vSelectName(lOptionIndex);
										_variant_t	vSelectIndex;

										// Get next option
										hr = spOptions->item(vSelectName, vSelectIndex, &spDisp);
										pOption = spDisp;

										if (SUCCEEDED(hr))
										{
											BSTR bsText = SysAllocString(g_szNullW);
											BSTR bsValue = SysAllocString(g_szNullW);

											// Get option text
											pOption->get_text(&bsText);
											// Get option value
											pOption->get_value(&bsValue);
											// Does option text/value = replacement text ?
											if ((bsText && (wcsicmp(bsText, wsText) == 0))
												||
												(bsValue && (wcsicmp(bsValue, wsText) == 0)))
											{
												// Select that option
												pOption->put_selected(TRUE);
												SysFreeString(bsText);
												SysFreeString(bsValue);
												break;
											}
											SysFreeString(bsText);
											SysFreeString(bsValue);
										}
									}
								}
							}
						}
					}
				}
			}
		}

		regkey.Close();	
	}

	return(bDone);
}

BOOL CFormWhizObj::TagMatch(_variant_t* vValue, CRegKey& regkey)
// This routine checks whether a given field name has a match
// with one of the user defined associations
{
	TCHAR 	szBuffer[MAX_FILL + MAX_TOKENS];
	TCHAR 	szValueName[10];
	TCHAR 	szDelim[] = _T(";");
	int		nIndex = 1;
	DWORD	dwBufSize = sizeof(szBuffer);
	BOOL	bDone = FALSE;
	long	lMax = 0;
	TCHAR	szTemp[MAX_FILL + 1];

	szTemp[0] = (TCHAR) '\0';
	_itot(nIndex, szValueName, 10);

	// Read each association
	while (regkey.QueryValue(szBuffer, szValueName, &dwBufSize) == ERROR_SUCCESS)
	{
		TCHAR	szFill[MAX_FILL];
		PTCHAR	pToken;
		long	lMatchLen;
		TCHAR	szPartialMatchChar[2];

		EncryptString(FALSE, szBuffer);

		// Get Field Data
		pToken = _tcstok(szBuffer, szDelim);

		if (pToken)
		{
			// Store replacement string
			_tcscpy(szFill, pToken);

			// Get Partial Match option			
			pToken = _tcstok(NULL, szDelim);
			if (pToken)
				_tcscpy(szPartialMatchChar, pToken);
			else
				return FALSE;
		}
		else
			return FALSE;

		if (pToken)
		{
			// Get field name tokens
			CRegExp	r;
			TCHAR	szTmpToken[MAX_TOKENS];
			
			while ((pToken = _tcstok(NULL, szDelim)) != NULL)
			{
				_tcsupr(pToken);
				
				if (szPartialMatchChar[0] != g_chPartialMatchOKChar)
				// Not a partial match, forces match with beginning and end of target string
				{
					// Modify regular expression
					_tcscpy(szTmpToken, "^");
					_tcscat(szTmpToken, pToken);
					_tcscat(szTmpToken, "$");
				}
				else
					_tcscpy(szTmpToken, pToken);
			
				// Token match field name?
				r.RegComp(szTmpToken);
				
				if (r.RegFind(_tcsupr((_bstr_t) vValue)) != -1)
				{
					lMatchLen = r.GetFindLen();
					if (lMatchLen > 0)
						bDone = TRUE;
					// Retain only longest match
					// First match with same length has priority
					if (bDone && (lMatchLen > lMax))
					{
						lMax = lMatchLen;
						_tcscpy(szTemp, szFill);
					}
				}
			}
		}
		nIndex++;
		_itot(nIndex, szValueName, 10);
		dwBufSize = sizeof(szBuffer);
	}

	if (bDone)	// Finally fill in form field
	{
		if (_tcsicmp(szTemp, g_szEmpty) == 0)
			szTemp[0] = (TCHAR) '\0';
		vValue->SetString(szTemp);
	}

	return bDone;
}

/////////////////////////////////////////////////////////////////////////////
// IEKeyboardProc - Keyboard hook procedure. It is necessarily a static 
// member of the class.
// 
LRESULT CALLBACK CFormWhizObj::IEKeyboardProc(int nCode, WPARAM wParam, LPARAM lParam)
{
	// We need a pointer to an instance of CFormWhizObj to work with. This value 
	// has been stored in the FwInstances collection when CFormWhizObj::SetSite has
	// been called.
	HWND hWnd = GetFocus();
	CFormWhizObj* pFw;

	// We first try to find a matching window
	do
	{
		pFw = LookupInstance(hWnd);
		if (pFw == NULL)
			hWnd = GetParent(hWnd);
	}
	while ((pFw == NULL) && (hWnd != NULL));

	// If the previous attempt failed, this may be because the IE window that
	// has the input focus is not a child of the main browser window. A typical 
	// case is the Favorites | Add Favorites | New Folder command. In this case,
	// we try to lookup using the current thread id
	if (pFw == NULL)
		pFw = LookupInstance(GetCurrentThreadId());

	if (pFw == NULL)	// Should not happen
		return 1;		// No other choice

	if (!pFw->m_bHotkeysEnabled || (nCode < 0) || (lParam & 0xC0000000))	// bit 30 or bit 31 is set
		// Process the key only once.
		// Disable hotkeys if still navigating.
		return CallNextHookEx(pFw->m_hHook, nCode, wParam, lParam);

	CRegKey regkey;

	// Get key values from registry
	if (regkey.Open(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
	{
		DWORD	dwKey;
		BOOL	bDone = FALSE;
		UINT	nState = pFw->GetKeyState();

		// Get value of hotkeys and compare with incoming keystroke
		if (pFw->m_bDocComplete
			&&
			(regkey.QueryValue(dwKey, g_szSerializeKey) == ERROR_SUCCESS)
			&&
			(bDone = (pFw->KeyMatch(nState, wParam, dwKey))))
			pFw->SerializeForms(TRUE);
		else if ((regkey.QueryValue(dwKey, g_szDeleteKey) == ERROR_SUCCESS)
				&&
				(bDone = (pFw->KeyMatch(nState, wParam, dwKey))))
			pFw->DeleteFormFile();
		else if ((regkey.QueryValue(dwKey, g_szConfigKey) == ERROR_SUCCESS)
				&&
				(bDone = (pFw->KeyMatch(nState, wParam, dwKey))))
			pFw->RunConfig();
		else if ((regkey.QueryValue(dwKey, g_szToggleKey) == ERROR_SUCCESS)
				&&
				(bDone = (pFw->KeyMatch(nState, wParam, dwKey))))
			pFw->ToggleState();
						
		regkey.Close();
		if (bDone)
			return 1;
	}

	// Note that we do not let the keystroke go through.
	// Which means that if the keystroke is used by MSIE
	// the corresponding function will not be triggered.
	// Use CallNextHook instead if you want that the original 
	// function be triggered as well.

	return CallNextHookEx(pFw->m_hHook, nCode, wParam, lParam);
}

UINT CFormWhizObj::GetKeyState()
// Check whether a key and a key option match
{
	UINT	nState = 0;
	
	// Make KeyState
	if (::GetKeyState(VK_SHIFT) < 0)
		nState += HOTKEYF_SHIFT;
	if (::GetKeyState(VK_MENU) < 0)
		nState += HOTKEYF_ALT;
	if (::GetKeyState(VK_CONTROL) < 0)
		nState += HOTKEYF_CONTROL;

	return(nState);	
}

BOOL CFormWhizObj::KeyMatch(UINT nState, WPARAM wParam, DWORD dwKey)
{
	return ((wParam == LOBYTE(dwKey)) && (nState == HIBYTE(dwKey)));	
}	

BOOL CFormWhizObj::SerializeForms(BOOL bStoring, IHTMLDocument2* pIDoc, BOOL* pbFormNotEmpty, BOOL* pbPwdRecorded)
// Store or read the contents of all document forms to a file
// The function is recursive: if pIDoc is NULL, this means 
// that we want to serialize all documents (typically the first
// call to SerializeForms). Then the function will be called for
// each document. pbPwdRecorded is a flag indicating whether
// a password field has been stored during serialization.
{
	CComPtr<IDispatch> spDisp;
	HRESULT hr;

	if (!g_bActive)
		return(FALSE);

	static BOOL bPwdRecorded;
	static BOOL bFormNotEmpty;
	
	if (!pIDoc)
	// Enter a loop calling SerializeForms for each known document
	{
		bPwdRecorded = FALSE;
		bFormNotEmpty = FALSE;

		for (UINT i = 0; i < m_nDocNum; i++)
			SerializeForms(bStoring, Documents[i], &bFormNotEmpty, &bPwdRecorded);

		if (bStoring && bFormNotEmpty)
		{
			FormWhizError(IDS_URLADDEDTEXT, errInfo);
			// Do we have stored a password?
			if (bPwdRecorded)
				FormWhizError(IDS_PWDWARNINGTEXT, errWarning);
		}
			
		return(TRUE);
	}

	// Start scanning of document
	if (pIDoc)
	{		
		TCHAR szURL[MAX_URL];

		// Get URL
		if (!RetrieveURL(pIDoc, szURL, MAX_URL - 1))
			return(FALSE);

		CComPtr<IHTMLElementCollection> pForms;
		long lFormNum, lFormIndex;

		CRegKey regkey;
		TCHAR	szFileName[MAX_PATH];
		TCHAR	szPathName[MAX_PATH];
		HANDLE	hFile;
		DWORD	dwSize = sizeof(szFileName);
		
		// Open or Create the URLs key
		if (regkey.Create(HKEY_CURRENT_USER, g_szURLKey) != ERROR_SUCCESS)
			return (FALSE);
		else
		{
			if (regkey.QueryValue(szFileName, szURL, &dwSize) != ERROR_SUCCESS)
			{
				if (bStoring)	// Store a form
				{
					// Compute filename and create file
					GetModuleFileName(GetModuleHandle(g_szModuleName), szPathName, MAX_PATH);
					*(_tcsrchr(szPathName, (TCHAR) '\\') + 1) = (TCHAR) '\0';
					_tcscat(szPathName, g_szData);

					// Try to create directory in any case, does nothing if directory exists
					CreateDirectory(szPathName, NULL);
					// Compute form file name from temporary filename
					GetTempFileName(szPathName, g_szNull, 0, szFileName);
					DeleteFile(szFileName);
					// Replace .tmp extension
					*(_tcsrchr(szFileName, (TCHAR) '.') + 1) = (TCHAR) '\0';
					_tcscat(szFileName, _T("auf"));
					hFile = CreateFile(szFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
					if (hFile != INVALID_HANDLE_VALUE)
					{
						// For each created form file, we have two registry entries:
						// 1. The title entry that points to the URL
						// 2. The URL entry that points to the filename
						
						regkey.SetValue(szFileName, szURL);
						
						// Now retrieve document title
						BSTR	bsTitle = SysAllocString(g_szNullW);
						TCHAR	szTitle[MAX_TITLE];

						_tcscpy(szTitle, g_szFWPrefix);
						if ((pIDoc->get_title(&bsTitle) != S_OK) || (bsTitle[0] == '\0') || !bsTitle)
							// Use URL if no title
							_uiscat(szTitle, szURL);
						else
						{
							// Strip trailing spaces
							while (bsTitle[0] && (bsTitle[wcslen(bsTitle) - 1] == ' '))
								bsTitle[wcslen(bsTitle) - 1] = '\0';

							TCHAR	szValue[MAX_OPTION_VALUE];
							TCHAR	szBufTitle[MAX_OPTION_VALUE];
							DWORD	dwSize = sizeof(szValue);

							_uiscpy(szBufTitle, szTitle);
							_uiscat(szBufTitle, bsTitle);
							if ((bsTitle[0] == '\0') || ((regkey.QueryValue(szValue, szBufTitle, &dwSize) == ERROR_SUCCESS)))
							{
								DWORD	dwIndex;
								WCHAR	szNewTitle[(sizeof(g_szHtmlEmptyTitle) + 4) / 2];
								WCHAR	szIndex[5];
								CRegKey	OptionKey;
								
								wcscpy(szNewTitle, g_szHtmlEmptyTitle);
								
								if (OptionKey.Open(HKEY_CURRENT_USER, g_szOptionKey) == ERROR_SUCCESS)
								{
									if (OptionKey.QueryValue(dwIndex, g_szEmptyTitleIndex) != ERROR_SUCCESS)
										dwIndex = 0;
											 								
									_itow(dwIndex, szIndex, 10);
									wcscat(szNewTitle, szIndex);

									OptionKey.SetValue(dwIndex + 1, g_szEmptyTitleIndex);
									OptionKey.Close();
								}
									
								SysReAllocString(&bsTitle, szNewTitle);
							}
								
							_uiscat(szTitle, bsTitle);
						}

						SysFreeString(bsTitle);

						regkey.SetValue(szURL, szTitle);
					}
					else
						goto exit1;
				}

				// Retrieve a form
				else
				{
exit1:
					regkey.Close();
					return(FALSE);
				}
			}
			else
			{
				// Open file anyway
				hFile = CreateFile(szFileName, 
								   bStoring ? GENERIC_WRITE : GENERIC_READ, 
								   0, 
								   NULL, 
								   OPEN_EXISTING,
								   FILE_ATTRIBUTE_NORMAL, 
								   NULL);
				if (hFile == INVALID_HANDLE_VALUE)
					// Delete key pointing to wrong file
					regkey.DeleteValue(szURL);
			}
			regkey.Close();
			if (hFile == INVALID_HANDLE_VALUE)
				return(FALSE);
		}

		// Get pointer to Forms collection
		hr = pIDoc->get_forms(&pForms);
		if (SUCCEEDED(hr))
		{
			// Get number of forms
			pForms->get_length(&lFormNum);

			if (bStoring && (lFormNum == 0))
			{
				// Eliminate NULL size file 
				CloseHandle(hFile);
				DeleteFile(szFileName);
				// Eliminate unnecessary registry entry
				if (regkey.Create(HKEY_CURRENT_USER, g_szURLKey) == ERROR_SUCCESS)
				{
					regkey.DeleteValue(szURL);
					regkey.Close();
				}
				// Nothing to do
				return(FALSE);				
			}
			
			// For each form
			for (lFormIndex = 0; lFormIndex < lFormNum; lFormIndex++)
			{
				CComQIPtr<IHTMLFormElement, &IID_IHTMLFormElement> spForm;
				spDisp = NULL;

				_variant_t	vName(lFormIndex);
				_variant_t	vIndex;

				// Get next form tag
				hr = pForms->item(vName, vIndex, &spDisp);
				spForm = spDisp;

				if (SUCCEEDED(hr))
				{
					// Get form name
					TCHAR		szFormName[50];
					TCHAR		szReadFormName[50];
					BSTR		bsName = SysAllocString(g_szNullW);

					spForm->get_name(&bsName);
					if (bsName)
						_uiscpy(szFormName, bsName);
					else
					{
						TCHAR	szIndex[3];
											
						_tcscpy(szFormName, g_szForm);
						_itot(lFormIndex, szIndex, 10);
						_tcscat(szFormName, szIndex);
					}
					SysFreeString(bsName);

					if (bStoring)
					{
						if (!WriteString(hFile, szFormName))
						{
							FormWhizError(IDS_ERROR_STORING, errError);
							goto exit2;
						}
					}
					else
					{
						// Read Form name from file and compare
						if (!ReadString(hFile, szReadFormName, sizeof(szReadFormName)))
						{
							FormWhizError(IDS_ERROR_READING, errError);
							goto exit2;
						}
						if (_tcsicmp(szFormName, szReadFormName))
						{
							// Forms don't have same name; terminate
							FormWhizError(IDS_ERROR_FORMHASCHANGED, errWarning);
exit2:
							CloseHandle(hFile);
							return(FALSE);
						}
					}

					CComQIPtr<IHTMLElementCollection, &IID_IHTMLElementCollection> spInputElems;
					spDisp = NULL;

					// Get element collection
					_variant_t vTagName(_T("INPUT"));
					hr = spForm->tags(vTagName, &spDisp);
					spInputElems = spDisp;

					if (SUCCEEDED(hr))
					{
						long lElemNum, lElemIndex;

						spInputElems->get_length(&lElemNum);
						// For each INPUT element
						for (lElemIndex = 0; lElemIndex < lElemNum; lElemIndex++)
						{
							CComQIPtr<IHTMLElement, &IID_IHTMLElement> spInputElem;
							spDisp = NULL;

							_variant_t	vElemName(lElemIndex);
							_variant_t	vElemIndex;

							// Get element
							hr = spInputElems->item(vElemName, vElemIndex, &spDisp);
							spInputElem = spDisp;

							if (SUCCEEDED(hr))
							{
								_variant_t		vValue(g_szNull);
								_variant_t		vType(g_szNull);
								_variant_t		vName(g_szNull);
								VARIANT_BOOL	bStatus;

								BSTR bsAttr = SysAllocString(g_szHtmlType);
								// Get element type
								spInputElem->getAttribute(bsAttr, FALSE, &vType);
								// No type means TEXT
								if (!(bstr_t) vType)
										vType.SetString(g_szHtmlText);
								SysFreeString(bsAttr);
								if (!(_tcsicmp((_bstr_t)vType, g_szHtmlText))
									||
									!(_tcsicmp((_bstr_t)vType, g_szHtmlPassword))
									||
									!(_tcsicmp((_bstr_t)vType, g_szHtmlRadio))
									||
									!(_tcsicmp((_bstr_t)vType, g_szHtmlCheckBox)))
								{
									bsAttr = SysAllocString(g_szHtmlName);
									// Get element name
									spInputElem->getAttribute(bsAttr, FALSE, &vName);
									if (!(bstr_t) vName)
										vName.SetString(g_szNull);
									SysReAllocString(&bsAttr, g_szHtmlValue);
									// Get element value
									spInputElem->getAttribute(bsAttr, FALSE, &vValue);
									if (!(bstr_t) vValue)
										vValue.SetString(g_szNull);
									SysFreeString(bsAttr);

									// If element is a checkbox or a radio button,
									// we should also store the status, otherwise
									// we store -1

									CComQIPtr<IHTMLOptionButtonElement, &IID_IHTMLOptionButtonElement> spOptionButtonElem;

									BOOL bStatusNeeded = !(_tcsicmp((_bstr_t)vType, g_szHtmlRadio))
															||
														 !(_tcsicmp((_bstr_t)vType, g_szHtmlCheckBox));
									if (bStatusNeeded)
									{
										spDisp = NULL;
										hr = spInputElems->item(vElemName, vElemIndex, &spDisp);
										spOptionButtonElem = spDisp;
										spOptionButtonElem->get_status(&bStatus);
									}

									BOOL bOK;
									
									if (bStoring)
									{
										// Store element name, type, value and status
										bOK = WriteString(hFile, (bstr_t) vName);
										bOK = bOK && WriteString(hFile, (bstr_t) vType);
										bOK = bOK && WriteString(hFile, (bstr_t) vValue);
										
										// Flag password recording
										bPwdRecorded = bPwdRecorded 
														||
													   (((LPCTSTR)(bstr_t) vValue)
													   		&&
														(((LPCTSTR)(bstr_t) vValue)[0] != '\0')
													   		&& 
														(!_tcsicmp((bstr_t) vType, g_szHtmlPassword)));
										
										int 	nStatus;
										DWORD	dwWritten;

										if (bStatusNeeded)
											nStatus = bStatus ? 1 : 0;
										else
											nStatus = -1;
										bOK = bOK && WriteFile(hFile, &nStatus, sizeof(int), &dwWritten, NULL);
										
										if (!bOK)
										{
											FormWhizError(IDS_ERROR_STORING, errError);
											CloseHandle(hFile);
											return(FALSE);
										}
										else
											bFormNotEmpty = TRUE;
									}
									else
									{
										TCHAR szReadName[MAX_NAME];
										TCHAR szReadType[MAX_NAME];
										TCHAR szReadValue[MAX_NAME];

										bOK = ReadString(hFile, szReadName, sizeof(szReadName));
										bOK = bOK && ReadString(hFile, szReadType, sizeof(szReadType));
										if (_tcsicmp(szReadName, (_bstr_t) vName) || _tcsicmp(szReadType, (_bstr_t) vType))
										{
											// if name and type don't match, exit 
											FormWhizError(IDS_ERROR_FORMHASCHANGED, errWarning);
											CloseHandle(hFile);
											return(FALSE);
										}
										bOK = ReadString(hFile, szReadValue, sizeof(szReadValue));
										vValue.SetString(szReadValue);
										// Read status string anyway
										int 	nStatus;
										DWORD	dwRead;

										bOK = bOK && ReadFile(hFile, &nStatus, sizeof(int), &dwRead, NULL);
										
										if (!bOK)
										{
											FormWhizError(IDS_ERROR_READING, errError);
											CloseHandle(hFile);
											return(FALSE);
										}

										// Change element value
										bsAttr = SysAllocString(g_szHtmlValue);
										spInputElem->setAttribute(bsAttr, vValue, FALSE);
										SysFreeString(bsAttr);

										if (bStatusNeeded)
										{
											if (nStatus != -1)
											{
												// Change element status
												bStatus = nStatus;
												spOptionButtonElem->put_status(bStatus);
											}
										}
									}
								}
							}
						}
					}

					// Now enumerate SELECT tags
					CComQIPtr<IHTMLElementCollection, &IID_IHTMLElementCollection> spSelectElems;
					spDisp = NULL;

					// Get select element collection
					_variant_t vSelectName(g_szHtmlSelect);
					hr = spForm->tags(vSelectName, &spDisp);
					spSelectElems = spDisp;

					if (SUCCEEDED(hr))
					{
						long lSelectNum, lSelectIndex;

						spSelectElems->get_length(&lSelectNum);
						// For each SELECT element
						for (lSelectIndex = 0; lSelectIndex < lSelectNum; lSelectIndex++)
						{
							CComQIPtr<IHTMLSelectElement, &IID_IHTMLSelectElement> spSelectElem;
							spDisp = NULL;

							_variant_t	vSelectName(lSelectIndex);
							_variant_t	vSelectIndex;

							// get next select element
							hr = spSelectElems->item(vSelectName, vSelectIndex, &spDisp);
							spSelectElem = spDisp;

							if (SUCCEEDED(hr))
							{
								BSTR	bsName = SysAllocString(g_szNullW);
								TCHAR	szElemName[MAX_NAME];

								// Get element name
								spSelectElem->get_name(&bsName);
								_uiscpy(szElemName, bsName);
								SysFreeString(bsName);
								
								if (bStoring)
								{
									if (!WriteString(hFile, szElemName))
										FormWhizError(IDS_ERROR_STORING, errError);
								}
								else
								{
									TCHAR	szName[MAX_NAME];
									BOOL	bOK;
									
									if (!(bOK = ReadString(hFile, szName, sizeof(szName)))
										||
										(_tcsicmp(szName, szElemName)))
									{
										if (!bOK)
											FormWhizError(IDS_ERROR_READING, errError);
										else
											FormWhizError(IDS_ERROR_FORMHASCHANGED, errWarning);
										CloseHandle(hFile);
										return(FALSE);
									}
								}
								
								CComQIPtr<IHTMLElementCollection, &IID_IHTMLElementCollection> spOptions;
								spDisp = NULL;
										
								_variant_t vOptionName(g_szHtmlOption);
							
								// Get option collection
								hr = spSelectElem->tags(vOptionName, &spDisp);
								spOptions = spDisp;

								if (SUCCEEDED(hr))
								{
									long lOptionNum, lOptionIndex;

									spOptions->get_length(&lOptionNum);
									for (lOptionIndex = 0; lOptionIndex < lOptionNum; lOptionIndex++)
									{
										CComQIPtr<IHTMLOptionElement, &IID_IHTMLOptionElement>pOption;
										spDisp = NULL;

										_variant_t	vSelectName(lOptionIndex);
										_variant_t	vSelectIndex;
	
										// Get next option									
										hr = spOptions->item(vSelectName, vSelectIndex, &spDisp);
										pOption = spDisp;

										if (SUCCEEDED(hr))
										{
											BSTR	bsText = SysAllocString(g_szNullW);
											BSTR	bsValue = SysAllocString(g_szNullW);
											TCHAR	szText[MAX_OPTION_TEXT];
											TCHAR	szValue[MAX_OPTION_VALUE];

											// Get option text
											pOption->get_text(&bsText);
											if (!bsText)
												SysReAllocString(&bsText, g_szNullW);
											_uiscpy(szText, bsText);
											SysFreeString(bsText);
											// Get option value
											pOption->get_value(&bsValue);
											if (!bsValue)
												SysReAllocString(&bsValue, g_szNullW);
											_uiscpy(szValue, bsValue);
											SysFreeString(bsValue);
									
											BOOL 		 bOK;
											VARIANT_BOOL bSelected;
													
											if (bStoring)
											{
												bOK = WriteString(hFile, szText);
												bOK = bOK && WriteString(hFile, szValue);

												DWORD	dwWritten;
												
												pOption->get_selected(&bSelected);
												bOK = bOK && WriteFile(hFile, &bSelected, sizeof(bSelected), &dwWritten, NULL);

												if (!bOK)
												{
													FormWhizError(IDS_ERROR_STORING, errError);
													CloseHandle(hFile);
													return(FALSE);
												}
											}
											else
											{
												TCHAR	szReadText[MAX_OPTION_TEXT];
												TCHAR	szReadValue[MAX_OPTION_VALUE];
												DWORD	dwRead;

												bOK = ReadString(hFile, szReadText, sizeof(szReadText));
												bOK = bOK && ReadString(hFile, szReadValue, sizeof(szReadValue));
												bOK = bOK && ReadFile(hFile, &bSelected, sizeof(bSelected), &dwRead, NULL);

												if (!bOK)
												{
													FormWhizError(IDS_ERROR_READING, errError);
													CloseHandle(hFile);
													return(FALSE);
												}
												
												// Do Text and Value match with read strings?
												// Was option selected when storing?
												if ((!_tcsicmp(szText, szReadText) || !_tcsicmp(szValue, szReadValue))
													 &&
													 bSelected)
													// Yes, select that option
													pOption->put_selected(TRUE);
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}

		CloseHandle(hFile);
	}

	return TRUE;
}

BOOL CFormWhizObj::WriteString(HANDLE hFile, LPCTSTR lpszString)
// Encrypt and write a string to form file
{
	DWORD	dwLen, dwWritten;
	BOOL	bOK;

	if (!lpszString)
		lpszString = g_szNull;

	TCHAR szBuffer[MAX_WRITE_STR];
	_tcscpy(szBuffer, lpszString);

	dwLen = (_tcslen(szBuffer) + 1) * sizeof(TCHAR);
	bOK = WriteFile(hFile, &dwLen, sizeof(DWORD), &dwWritten, NULL);
	// Encrypt string (no real encryption here - 
	// just to prevent data from  appearing in clear text)
	EncryptString(TRUE, szBuffer);
	bOK = bOK && WriteFile(hFile, szBuffer, dwLen, &dwWritten, NULL);

	return(bOK && (dwWritten == dwLen));
}

BOOL CFormWhizObj::ReadString(HANDLE hFile, LPSTR lpszString, DWORD dwMaxLen)
// Read and decrypt a string from a form file
{
	DWORD	dwLen, dwRead;
	BOOL	bOK;

	bOK = ReadFile(hFile, &dwLen, sizeof(DWORD), &dwRead, NULL);
	if ((dwRead == sizeof(DWORD)) && (dwLen <= dwMaxLen))
	{
		bOK = bOK && ReadFile(hFile, lpszString, dwLen, &dwRead, NULL);
		// Decrypt string
		EncryptString(FALSE, lpszString);
		return(bOK && (dwRead == dwLen));
	}
	else
		return(FALSE);
}

void CFormWhizObj::DeleteFormFile()
// Destroy a form file and update registry
{
	if (!g_bActive)
		return;

	CRegKey regkey;
	BOOL bDone = FALSE;

	if (regkey.Open(HKEY_CURRENT_USER, g_szURLKey) != ERROR_SUCCESS)
		return;

	// For all documents
	for (UINT i = 0; i < m_nDocNum; i++)
	{
		TCHAR szURL[MAX_URL];
		// Check URL
		if (!RetrieveURL(Documents[i], szURL, MAX_URL - 1))
			break;

		// Delete form file and registry key
		TCHAR	szFileName[MAX_PATH];
		DWORD	dwSize = sizeof(szFileName);

		if (regkey.QueryValue(szFileName, szURL, &dwSize) == ERROR_SUCCESS)
		{
			DeleteFile(szFileName);
			regkey.DeleteValue(szURL);
			bDone = TRUE;
		}

		// Now retrieve document title
		BSTR	bsTitle = SysAllocString(g_szNullW);
		TCHAR	szTitle[MAX_TITLE];
		
		_tcscpy(szTitle, g_szFWPrefix);
		if ((Documents[i]->get_title(&bsTitle) != S_OK) || (bsTitle[0] == '\0') || !bsTitle)
			// Use URL if no title
			_uiscat(szTitle, szURL);
		else
		{
			while (bsTitle[0] && (bsTitle[wcslen(bsTitle) - 1] == ' '))
				bsTitle[wcslen(bsTitle) - 1] = '\0';

			if (bsTitle[0] == '\0')
				SysReAllocString(&bsTitle, g_szHtmlEmptyTitle);
			_uiscat(szTitle, bsTitle);
		}

		SysFreeString(bsTitle);
		regkey.DeleteValue(szTitle);
	}

	regkey.Close();
	
	if (bDone)
		FormWhizError(IDS_DELKEYTEXT, errInfo);
}

BOOL CFormWhizObj::RetrieveURL(IHTMLDocument2* pIDoc, LPSTR lpszURL, UINT nMaxLen)
// Given a IHTMLDocument2*, get URL string
{
	if (!lpszURL)
		return(FALSE);
		
	// Get and return URL for specified document
	BOOL bOK;
	BSTR bsURL = SysAllocString(g_szNullW);

	pIDoc->get_URL(&bsURL);
	if ((bsURL) && (wcslen(bsURL) <= nMaxLen))
	{
		_uiscpy(lpszURL, bsURL);
		bOK = TRUE;
	}
	else
	{
		*lpszURL = (TCHAR) '\0';
		bOK = FALSE;
	}

	SysFreeString(bsURL);
	
	return(bOK);
}		

void CFormWhizObj::AddDocument(IHTMLDocument2* pDoc)
// Add a document to our list
{
	UINT i;
	
	if (!pDoc)
		return;

	// No duplicates
	for (i = 0; i < m_nDocNum; i++)
		if (Documents[i] == pDoc)
			return;

	if (m_nDocNum < MAX_DOCUMENTS)
	{
		Documents[m_nDocNum] = pDoc;
		m_nDocNum++;
	}
}

void CFormWhizObj::RemoveDocument(IHTMLDocument2* pDoc)
// Remove a document from list
{
	UINT i;

	// Find and remove specified document object
	for (i = 0; i < m_nDocNum; i++)
		if (Documents[i] == pDoc)
		{
			if (i < MAX_DOCUMENTS - 1)
				memmove(Documents[i], Documents[i+1], sizeof(IHTMLDocument2*) * (m_nDocNum - i - 1));
			m_nDocNum--;
			break;
		}
	
	return;
}

void CFormWhizObj::RunConfig()
// Run FormWhiz configurator
{
	TCHAR	szPathName[MAX_PATH];

	// Compute path of configuration program
	GetModuleFileName(GetModuleHandle(g_szModuleName), szPathName, MAX_PATH);
	*(_tcsrchr(szPathName, (TCHAR) '\\') + 1) = (TCHAR) '\0';
	_tcscat(szPathName, g_szConfName);
	
	WinExec(szPathName, SW_SHOWNORMAL);
}

void CFormWhizObj::ToggleState()
// Toggle active and inactive state
{
	if (g_bToggleInUse)
		return;

	g_bActive = !g_bActive;

	TCHAR szText[MAX_MSG_TEXT];
	TCHAR szCaption[MAX_MSG_CAPTION];

	LoadString(_Module.m_hInst, IDS_ACTIVECAPTION + (g_bActive ? 0 : 2), szCaption, sizeof(szCaption));
	LoadString(_Module.m_hInst, IDS_ACTIVETEXT + (g_bActive ? 0 : 2), szText, sizeof(szText));

	g_bToggleInUse = TRUE;
	MessageBox(m_hwndIE, szText, szCaption, MB_OK | MB_ICONINFORMATION);
	g_bToggleInUse = FALSE;
}

void CFormWhizObj::FormWhizError(int nTextId, errType e)
{
	TCHAR	szText[MAX_MSG_TEXT];
	TCHAR	szCaption[MAX_MSG_CAPTION];
	UINT	nFlag;
	UINT	nCaptionId;

	switch(e)
	{
		case errWarning:
			nCaptionId = IDS_FORMWHIZWARNINGCAPTION;
			nFlag = MB_ICONWARNING;
			break;

		case errError:
			nCaptionId = IDS_FORMWHIZERRORCAPTION;
			nFlag = MB_ICONEXCLAMATION;
			break;
	
		case errInfo:
			nCaptionId = IDS_FORMWHIZINFOCAPTION;
			nFlag = MB_ICONINFORMATION;
			break;
	}

	LoadString(_Module.m_hInst, nCaptionId, szCaption, sizeof(szCaption));
	LoadString(_Module.m_hInst, nTextId, szText, sizeof(szText));
	MessageBox(m_hwndIE, szText, szCaption, MB_OK | nFlag);
}